package rbas.util;

import edu.cmu.tetrad.util.MatrixUtils;
import edu.cmu.tetrad.util.ProbUtils;

import java.util.Random;

/**
 * Complements the ProbUtils class from Tetrad.
 *
 * @author: Ricardo Silva
 * @version $Revision: 1.1 $ $Date: 2005/08/30 20:29:17 $
 */

public class RBASProbUtils {


    /**
     * Bivariate normal CDF. Adapted from statlib, file general/gaut.c, by Ajay
     * Shah. Adaptation for Java found in http://www.geocities.com/optionpage/utility.html
     * --November 1st 2003, Ricardo Silva
     *
     * @param ah upper bound 1
     * @param ak upper bound 1
     * @param r  correlation
     * @return Prob(x1 <= ah, x2 <= ak)
     */

    public static double biNormalCdf(double ah, double ak, double r) {
        return biNormalCdf2(-ah, -ak, r);
    }

    private static double biNormalCdf2(double ah, double ak, double r) {
        double a2, ap, b, cn, conex, ex, g2, gh, gk, gw = 0, h2, h4, rr, s1, s2, sgn, sn, sp, sqr, t, temp, w2, wh = 0, wk = 0;
        int is = 0;
        int myflag = 1;
        final double con = (2.0 * Math.PI / 2.0) * 0.0000000001;// 10.0e-10;
        temp = -ah;
        gh = ProbUtils.normalCdf(temp);
        gh = gh / 2.0;
        temp = -ak;
        gk = ProbUtils.normalCdf(temp);
        gk = gk / 2.0;

        b = 0;


        if (r == 0) {
            b = 4 * gh * gk;
        }
        else {
            rr = 1 - r * r;
            assert rr >= 0;
            if (rr != 0) {
                sqr = Math.sqrt(rr);
                if (ah == 0) {
                    if (ak == 0) {
                        b = Math.atan(r / sqr) / (2 * Math.PI) + .25;
                        if (b < 0) {
                            b = 0;
                        }
                        if (b > 1) {
                            b = 1;
                        }
                        return (b);
                    }
                    b = b + gk;
                }
                else {
                    b = gh;
                    if (ah * ak != 0) {
                        if (ah * ak < 0) {
                            b = b - .5;
                        }
                        b = b + gk;
                    }
                    wh = -ah;
                    wk = (ak / ah - r) / sqr;
                    gw = 2 * gh;
                    is = -1;
                    myflag = -1;
                }
                do {
                    if (myflag == -1) {
                        myflag = 1;
                    }
                    else {
                        wh = -ak;
                        wk = (ah / ak - r) / sqr;
                        gw = 2 * gk;
                        is = 1;
                    }
                    sgn = -1;
                    t = 0;
                    if (wk != 0) {
                        if (Math.abs(wk) != 1) {
                            if (Math.abs(wk) > 1) {
                                sgn = -sgn;
                                wh = wh * wk;
                                g2 = ProbUtils.normalCdf(wh);
                                wk = 1 / wk;
                                if (wk < 0) {
                                    b = b + .5;
                                }
                                b = b - (gw + g2) / 2 + gw * g2;
                            }
                            h2 = wh * wh;
                            a2 = wk * wk;
                            h4 = h2 * .5;
                            ex = 0;
                            if (h4 < 150.0) {
                                ex = Math.exp(-h4);
                            }
                            w2 = h4 * ex;
                            ap = 1;
                            s2 = ap - ex;
                            sp = ap;
                            s1 = 0;
                            sn = s1;
                            conex = Math.abs(con / wk);
                            do {
                                cn = ap * s2 / (sn + sp);
                                s1 = s1 + cn;
                                if (Math.abs(cn) <= conex) {
                                    break;
                                }
                                sn = sp;
                                sp = sp + 1;
                                s2 = s2 - w2;
                                w2 = w2 * h4 / sp;
                                ap = -ap * a2;
                            } while (true);
                            t = (Math.atan(wk) - wk * s1) / (2 * Math.PI);
                        }
                        else {
                            t = wk * gw * (1 - gw) / 2;
                        }

                        b = b + sgn * t;
                    }
                    if (is >= 0) {
                        break;
                    }
                } while (ak != 0);
            }
            else if (r >= 0) {
                if (ah >= ak) {
                    b = 2 * gh;
                }
                else {
                    b = 2 * gk;
                }
            }
            else if (ah + ak < 0) {
                b = 2 * (gh + gk) - 1;
            }
        }
        if (b < 0) {
            b = 0;
        }
        if (b > 1) {
            b = 1;
        }
        return (b);
    }


    /**
     * Compute the probability over a rectangular region with correlation matrix
     * c. Algorithm extracted from </p> Alan Genz: Numerical Computation of
     * Multivariate Normal Probabilities; revised version published in J. Comp.
     * Graph Stat. 1 (1992), pp. 141-149. http://www.math.wsu.edu/faculty/genz/homepage
     * </p> Warning: this method has the side effect of changing the order of
     * the elements in the arrays given as input. </p> --November 3st 2003,
     * Ricardo Silva
     *
     * @param a lower bounds (use Double.NEGATIVE_INFINITY if necessary)
     * @param b upper bounds (use Double.POSITIVE_INFINITY if necessary)
     */

    public static double multinormalProb(double a[], double b[], double cov[][]) {
        assert a.length == b.length;
        for (int i = 0; i < a.length; i++) {
            assert a[i] <= b[i];
        }
        /*for (int i = 0; i < a.length; i++)
            System.out.println("a = " + a[i] + ", b = " + b[i]);
        for (int i = 0; i < cov.length; i++) {
             for (int j = 0; j < cov.length; j++)
                 System.out.print(cov[i][j] + "\t");
            System.out.println();
        }
        System.out.println();*/
        orderIntegral(a, b, cov);
        /*for (int i = 0; i < a.length; i++)
            System.out.println("a = " + a[i] + ", b = " + b[i]);
        for (int i = 0; i < cov.length; i++) {
             for (int j = 0; j < cov.length; j++)
                 System.out.print(cov[i][j] + "\t");
            System.out.println();
        }
        System.exit(0);*/
        double c[][] = MatrixUtils.cholesky(cov);
        double d[] = new double[a.length], e[] = new double[a.length], f[] = new double[a.length];
        double w[] = new double[a.length - 1], y[] = new double[a.length - 1];
        double intSum = 0., varSum = 0., error, auxSum;
        int n = 0;
        Random r = new Random();

        if (a[0] == Double.NEGATIVE_INFINITY) {
            d[0] = 0.;
        }
        else {
            d[0] = ProbUtils.normalCdf(a[0] / c[0][0]);
        }
        if (b[0] == Double.POSITIVE_INFINITY) {
            e[0] = 1.;
        }
        else {
            e[0] = ProbUtils.normalCdf(b[0] / c[0][0]);
        }
        f[0] = e[0] - d[0];

        do {
            for (int i = 0; i < w.length; i++) {
                w[i] = r.nextDouble();
            }
            for (int i = 1; i < a.length; i++) {
                double quant = d[i - 1] + w[i - 1] * (e[i - 1] - d[i - 1]);
                if (quant == 1.) {
                    y[i - 1] = Double.MAX_VALUE;
                }
                else if (quant == 0.) {
                    y[i - 1] = -Double.MAX_VALUE;
                }
                else {
                    y[i - 1] = ProbUtils.normalQuantile(d[i - 1] +
                            w[i - 1] * (e[i - 1] - d[i - 1]));
                }
                auxSum = 0.;
                for (int j = 0; j < i; j++) {
                    auxSum += c[i][j] * y[j];
                }
                if (a[i] == Double.NEGATIVE_INFINITY) {
                    d[i] = 0.;
                }
                else {
                    d[i] = ProbUtils.normalCdf((a[i] - auxSum) / c[i][i]);
                }
                if (b[i] == Double.POSITIVE_INFINITY) {
                    e[i] = 1.;
                }
                else {
                    e[i] = ProbUtils.normalCdf((b[i] - auxSum) / c[i][i]);
                }
                f[i] = (e[i] - d[i]) * f[i - 1];
            }
            intSum += f[f.length - 1];
            varSum += f[f.length - 1] * f[f.length - 1];
            n++;
            error = 2.5 * Math.sqrt((varSum / n - (intSum / n) * (intSum / n)) / n);
            //error = 2;
        } while ((error > 0.0001 && n < 5000) || n < 50);
        return intSum / n;
    }

    private static void orderIntegral(double a[], double b[], double cov[][]) {
        double temp1, temp2, temp3, temp4;
        int idx1, idx2;
        for (int i = 0; i < a.length - 1; i++) {
            int smallest = i;
            for (int j = i + 1; j < a.length; j++) {
                if (Double.isInfinite(a[j]) && Double.isInfinite(a[smallest])) {
                    if (b[j] < b[smallest]) {
                        smallest = j;
                    }
                }
                else if (Double.isInfinite(b[j]) &&
                        Double.isInfinite(b[smallest])) {
                    if (a[j] > a[smallest]) {
                        smallest = j;
                    }
                }
                else if (Double.isInfinite(a[smallest]) &&
                        Double.isInfinite(b[j])) {
                    if (b[smallest] <= 0. && a[j] >= 0. && a[j] > -b[smallest]) {
                        smallest = j;
                    }
                    else if (b[smallest] >= 0. && a[j] >= 0.) {
                        smallest = j;
                    }
                    else if (b[smallest] >= 0. && a[j] <= 0. &&
                            -a[j] < b[smallest]) {
                        smallest = j;
                    }
                }
                else if (Double.isInfinite(b[smallest]) &&
                        Double.isInfinite(a[j])) {
                    if (a[smallest] <= 0. && b[j] >= 0. && b[j] < -a[smallest]) {
                        smallest = j;
                    }
                    else if (a[smallest] <= 0. && b[j] <= 0.) {
                        smallest = j;
                    }
                    else if (a[smallest] >= 0. && b[j] <= 0. &&
                            -b[j] > a[smallest]) {
                        smallest = j;
                    }
                }
                else if ((Double.isInfinite(a[smallest]) ||
                        Double.isInfinite(b[smallest])) &&
                        !Double.isInfinite(a[j]) && !Double.isInfinite(b[j])) {
                    smallest = j;
                }
                else if (!Double.isInfinite(a[smallest]) &&
                        !Double.isInfinite(b[smallest]) &&
                        !Double.isInfinite(a[j]) && !Double.isInfinite(b[j]) && Math.abs(
                                b[j] - a[j]) <
                        Math.abs(b[smallest] - a[smallest])) {
                    smallest = j;
                }
            }
            temp1 = a[i];
            temp2 = b[i];
            a[i] = a[smallest];
            b[i] = b[smallest];
            a[smallest] = temp1;
            b[smallest] = temp2;
            if (smallest > i) {
                idx1 = smallest;
                idx2 = i;
            }
            else if (smallest < i) {
                idx1 = i;
                idx2 = smallest;
            }
            else {
                continue;
            }
            temp1 = cov[idx1][idx2];
            temp3 = cov[idx1][idx1];
            temp4 = cov[idx2][idx2];
            for (int j = 0; j < cov.length; j++) {
                temp2 = cov[idx1][j];
                cov[j][idx1] = cov[idx1][j] = cov[idx2][j];
                cov[j][idx2] = cov[idx2][j] = temp2;
            }
            cov[idx2][idx1] = cov[idx1][idx2] = temp1;
            cov[idx1][idx1] = temp4;
            cov[idx2][idx2] = temp3;
        }
    }

    /**
     * Digamma function. Adapted from clsMathParser, a Visual Basic package
     * of mathematical functions.
     *
     * @param x
     * @return
     */

    public static double digamma(double x) {
        double b1[] = new double[12], b2[] = new double[12];
        double z, s, tmp;
        int    LIM_LOW = 8;
        b1[0] = 1; b2[0] = 1;
        b1[1] = 1; b2[1] = 6;
        b1[2] = -1; b2[2] = 30;
        b1[3] = 1; b2[3] = 42;
        b1[4] = -1; b2[4] = 30;
        b1[5] = 5; b2[5] = 66;
        b1[6] = -691; b2[6] = 2730;
        b1[7] = 7; b2[7] = 6;
        b1[8] = -3617; b2[8] = 360;
        b1[9] = 43867; b2[9] = 798;
        b1[10] = -174611; b2[10] = 330;
        b1[11] = 854513; b2[11] = 138;
        if (x <= LIM_LOW)
            z = x - 1 + LIM_LOW;
        else
            z = x - 1;
        s = 0;
        for (int k = 1; k <= 11; k++) {
            tmp = b1[k] / b2[k] / k / Math.pow(z, (2 * k));
            s += tmp;
        }

        double y = Math.log(z) + 0.5 * (1 / z - s);

        if (x <= LIM_LOW) {
            s = 0;
            for (int k = 0; k < LIM_LOW; k++)
                s = s + 1 / (x + k);
            y = y - s;
        }

        return y;
    }

    /**
     * Trigamma function. Adapted from Schneider (Applied Statistics 27, 1978, p. 97-99).
     *
     * @param x
     * @return
     */

    public static double trigamma(double x) {
        double a = 1.e-4, b = 5.;
        double b2 = 0.1666666667, b4 = -0.03333333333, b6 = 0.2380952381, b8 = -0.03333333333;
        double trigam = 0.0;
        if (x <= 0)
            throw new RuntimeException("Invalid argument: trigamma function");
        double z = x;
        if (z <= a)
            return 1. / (z * z);
        while (z < b) {
            trigam += 1. / (z * z);
            z++;
        }
        double y = 1. / (z * z);
        trigam += 0.5 * y + (1. + y *(b2 + y * (b4 + y * (b6 + y * b8)))) / z;
        return trigam;
    }

    /**
     * Inverse of a digamma function. From Tom Minka's "Estimating a Dirichlet distribution".
     *
     * @param x
     * @return
     */

    public static double inverseDigamma(double x) {
        double current;
        if (x >= -2.22)
            current = Math.exp(x) + 0.5;
        else
            current = -1. / (x - digamma(1));
        for (int i = 0; i < 5; i++)
            current -= (digamma(current) - x) / trigamma(current);
        return current;
    }

    public static double klDivergenceDirichlet(double alpha1[], double alpha2[]) {
        double alpha1t = 0., alpha2t = 0.;
        for (int i = 0; i < alpha1.length; i++) {
            alpha1t += alpha1[i];
            alpha2t += alpha2[i];
        }
        double distance = ProbUtils.lngamma(alpha1t) - ProbUtils.lngamma(alpha2t);
        for (int i = 0; i < alpha1.length; i++)
            distance += ProbUtils.lngamma(alpha2[i]) - ProbUtils.lngamma(alpha1[i]) +
                        (alpha1[i] - alpha2[i]) * (RBASProbUtils.digamma(alpha1[i]) - RBASProbUtils.digamma(alpha1t));
        return distance;
    }

    public static double klDivergenceNormal(double mean1[], double cov1[][], double mean2[], double cov2[][]) {
        double meanDiff[] = MatrixUtils.subtract(mean1, mean2);
        double cov2Inv[][] = MatrixUtils.inverse(cov2);
        return  0.5 * (
                    Math.log(MatrixUtils.determinant(cov2) / MatrixUtils.determinant(cov1))
                  + MatrixUtils.trace(MatrixUtils.product(cov2Inv, cov1))
                  + MatrixUtils.innerProduct(meanDiff, MatrixUtils.product(cov2Inv, meanDiff))
                  - meanDiff.length);
    }

    public static double klDivergenceGamma(double a1, double b1, double a2, double b2) {
        return a1 * Math.log(b1) - a2 * Math.log(b2) - ProbUtils.lngamma(a1) + ProbUtils.lngamma(a2)
               + (a1 - a2) * (RBASProbUtils.digamma(a1) - Math.log(b1)) - a1 * (1. - b2 / b1);
    }

    public static double klDivergenceBeta(double a1, double b1, double a2, double b2) {
        double dg = RBASProbUtils.digamma(a1 + b1);
        return ProbUtils.lngamma(a1 + b1) - ProbUtils.lngamma(a2 + b2) -
               (ProbUtils.lngamma(a1) - ProbUtils.lngamma(a2) - (a1 - a2) * (RBASProbUtils.digamma(a1) - dg)) -
               (ProbUtils.lngamma(b1) - ProbUtils.lngamma(b2) - (b1 - b2) * (RBASProbUtils.digamma(b1) - dg));
    }

    public static double gaussianEntropy(double covMatrix[][]) {
        return 0.5 * (covMatrix.length * Math.log(2 * Math.PI) + covMatrix.length +
                     Math.log(MatrixUtils.determinant(covMatrix)));
    }

}


